/* $Id: array.c,v 1.8 2002/11/18 09:23:54 aa5779 Exp $ */
/*
	Copyright (C) 2001, 2002 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <string.h>
#include <stdlib.h>
#include "util.h"
#include "dstring.h"
#include "interp.h"
#define ARRAY_IMP
#include "array.h"

#define AB_SIZE 128
typedef struct array_block
{
	int bsize;
	dstring **dss;
} array_block;

typedef struct array_header
{
	int cursize;
	int block_cnt;
	array_block *blocks;
} array_header;

storage_class array_storage_class = {array_create, array_destroy, 
									 array_check, array_get,
									 array_put, array_add,
									 array_remove,
									 array_first, array_next,
									 array_enditer,
									 array_numofitems,
									 array_seek,
									 array_clone
									};

storage *array_create(interpreter *i, char *name, char *arg)
{
	storage *ns = ALLOC(storage);
	array_header *h = ALLOC(array_header);
	ns->data = h;
	ns->class = &array_storage_class;
	return ns;
}

void array_destroy(interpreter *i, storage *s)
{
	array_header *h = s->data;
	array_block *bi;
	dstring **di;

	for(bi = h->blocks; bi < h->blocks + h->block_cnt; bi++)
	{
		for(di = bi->dss; di < bi->dss + bi->bsize; di++)
		{
			if(*di)
				ds_unfix(*di);
		}
		if(bi->dss)
			free(bi->dss);
	}
	if(h->blocks)
		free(h->blocks);
	FREE(h);
}

static dstring **get_block(array_header *h, int idx)
{
	array_block *bi;
	div_t d;
	if(idx < 0)
		return NULL;
	d = div(idx, AB_SIZE);
	if(d.quot >= h->block_cnt)
		return NULL;
	bi = h->blocks + d.quot;
	return d.rem >= bi->bsize ? NULL : bi->dss + d.rem;
}

static dstring *get_value(array_header *h, int idx)
{
	dstring **dss = get_block(h, idx);
	return dss ? *dss : NULL;
}

int array_check(interpreter *i, storage *s, dstring *key)
{
	return get_value(s->data, convert_to_int(DS_BODY(key))) != NULL;
}

dstring *array_get(interpreter *i, storage *s, dstring *key)
{
	dstring* ds =  get_value(s->data, convert_to_int(DS_BODY(key)));
	if(!ds)
		recover_error("invalid array index: %s", DS_BODY(key));
	return ds;
}

void array_put(interpreter *i, storage *s, dstring *key, dstring *val)
{
	dstring* ds =  get_value(s->data, convert_to_int(DS_BODY(key)));
	if(!ds)
		recover_error("invalid array index: %s", DS_BODY(key));
	ds_assign(ds, val);
}

static dstring *expand_array(array_header *h, int idx)
{
	div_t d;
	array_block *b;
	if(idx < 0)
		recover_error("negative index: %d", idx);
	d = div(idx, AB_SIZE);
	if(d.quot >= h->block_cnt)
	{
		int temp = h->block_cnt;
		h->block_cnt = d.quot + 1;
		h->blocks = xrealloc(h->blocks, sizeof(*h->blocks) * h->block_cnt);
		memset((void *)(h->blocks + temp), 0, sizeof(*h->blocks) * (h->block_cnt - temp));
	}
	b = h->blocks + d.quot;
	if(d.rem >= b->bsize)
	{
		int temp = b->bsize;
		b->bsize  = d.rem + 1;
		b->dss = xrealloc(b->dss, sizeof(dstring *) * b->bsize);
		memset((void *)(b->dss + temp), 0, sizeof(*b->dss) * (b->bsize - temp));
	}
	if(!b->dss[d.rem])
	{
		h->cursize++;
		b->dss[d.rem] = ds_fix(ds_create(NULL));
	}
	return b->dss[d.rem];
}

void array_add(interpreter *i, storage *s, dstring *key, dstring *val)
{
	dstring* ds = expand_array(s->data, convert_to_int(DS_BODY(key)));
	ds_assign(ds, val);
}

void array_remove(interpreter *i, storage *s, dstring *key)
{
	dstring **dss = get_block(s->data, convert_to_int(DS_BODY(key)));
	if(!dss || !*dss)
		recover_error("invalid index: %s", DS_BODY(key));
	ds_unfix(*dss);
	((array_header *)s->data)->cursize--;
	*dss = NULL;
}

typedef struct array_iterator
{
	int bi;
	int offs;
	int idx;
} array_iterator;

static int find_nextindex(array_block *start, int offs, int *eoffs, array_block **end, int idx,
						  array_block *lastblock
						 )
{
	while(start <= lastblock)
	{
		while(offs < start->bsize)
		{
			if(start->dss[offs])
			{
				*end = start;
				*eoffs = offs;
				return idx;
			}
			offs++;
			idx++;
		}
		idx += AB_SIZE - offs;
		offs = 0;
		start++;
	}
	return -1;
}

dstring *array_first(interpreter *i, storage *s, void **iter)
{
	array_header *h = s->data;
	array_iterator *na;
	array_block *bi;
	if(!h->cursize)
		return NULL;
	na = ALLOC(array_iterator);
	na->idx = find_nextindex(h->blocks, 0, &na->offs, &bi, 0, h->blocks + h->block_cnt - 1);
	na->bi = bi - h->blocks;
	*iter = na;
	return ds_fromint(na->idx, 10, 0);
}

dstring *array_next(interpreter *i, storage *s, void **iter)
{
	array_header *h = s->data;
	array_iterator *it = *iter;
	array_block *bi;
	it->idx = find_nextindex(h->blocks + it->bi, it->offs + 1, &it->offs, &bi, it->idx + 1,
							 h->blocks + h->block_cnt - 1);
	if(it->idx < 0)
	{
		FREE(it);
		*iter = NULL;
		return NULL;
	}
	it->bi = bi - h->blocks;
	return ds_fromint(it->idx, 10, 0);
}

void array_enditer(interpreter *i, storage *s, void **iter)
{
	if(*iter)
	{
		FREE((array_iterator *)*iter);
		*iter = NULL;
	}
}

int array_numofitems(interpreter *i, storage *s)
{
	return ((array_header *)s->data)->cursize;
}

dstring *array_seek(interpreter *i, storage *s, void **iter, dstring *key)
{
	recover_error("not yet implemented");
	return NULL; // never reached
}

storage *array_clone(storage *src)
{
	storage *dest = ALLOC(storage);
	array_header *old = src->data, *nh = ALLOC(array_header);
	array_block *iter, *iter1;
	dstring **siter, **siter1;

	dest->class = src->class;
	dest->data = nh;
	nh->cursize = old->cursize;
	nh->block_cnt = old->block_cnt;
	iter1 = nh->blocks = xcalloc(nh->block_cnt, sizeof(nh->blocks));
	for(iter = old->blocks; iter < old->blocks + old->block_cnt; iter++, iter1++)
	{
		if((iter1->bsize = iter->bsize))
		{
			siter1 = iter1->dss = xcalloc(iter1->bsize, sizeof(*iter1->dss));
			for(siter = iter->dss; siter < iter->dss + iter->bsize; siter++, siter1++)
			{
				if(*siter)
					*siter1 = ds_fix(ds_copy(*siter));
			}
		}
	}
	return dest;
}
