/* $Id: table.c,v 1.12 2003/01/02 22:41:14 aa5779 Exp $ */
/*
	Copyright (C) 2001, 2002 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#include <limits.h>
#include DB_H
#include <errno.h>
#include "util.h"
#include "dstring.h"
#include "lists.h"
#include "interp.h"
#define TABLE_IMP
#include "table.h"


storage_class table_storage_class = {table_create, table_destroy, 
									 table_check, table_get,
									 table_put, table_add,
									 table_remove,
									 table_first, table_next,
									 table_enditer,
									 table_numofitems,
									 table_seek
									};

storage_class hash_storage_class = {hash_create, table_destroy, 
									 table_check, table_get,
									 table_put, table_add,
									 table_remove,
									 table_first, table_next,
									 table_enditer,
									 table_numofitems,
									 table_seek
									};

typedef DB *db_t;
typedef DBT dbdata_t;

static void db_error(void)
{
	recover_error("DB error: %s", strerror(errno));
}

DEFLIST(dbf_list,
		db_t db;
		int fd;
		int reccount;
		unsigned char is_hash : 1;
		unsigned char traverse_hash : 1;
		unsigned char reversed : 1;
		unsigned char read_only : 1;
		);

static int is_always_new;
static int read_only;
static int table_mode;

static void *open_db(const char *fname)
{
	return dbopen(fname, (is_always_new ? O_TRUNC : 0) |
						 (read_only ? O_RDONLY : O_RDWR | O_CREAT),
						 S_IREAD | S_IWRITE, table_mode, NULL);
}
	
static dbf_list *all_tables;

static storage *itable_create(int mode, interpreter *i, char *name, char *arg)
{
	storage *ns = ALLOC(storage);
	db_t db;
	dbf_list *ndbl;
	char *primary_env = NULL;
	int status;
	dbdata_t ky, val;

   	is_always_new = (*arg == '!');
	read_only = (*arg == '+');
	table_mode = mode;
	arg += is_always_new || read_only;
	if(is_always_new)
		primary_env = getenv(SL_TABLE_NEW_PATH_ENV);
	else if(read_only)
		primary_env = getenv(SL_TABLE_RO_PATH_ENV);
	if(!primary_env)
		primary_env = getenv(SL_TABLE_PATH_ENV);
	ndbl = ALLOC(dbf_list);
	db = *arg ? find_filename(arg, primary_env, "t", open_db, open_db) : open_db(NULL);
	if(!db)
		recover_error("can't load DB '%s' (%s): %s", name, arg, strerror(errno));
	ns->data = ndbl;
	ns->class = &table_storage_class;
	ndbl->db = db;
	ndbl->read_only = read_only;
	ndbl->is_hash = mode == DB_HASH;
#ifdef CAN_LOCK	
	if(!*arg)
		ndbl->fd = -1;
	else
	{
		struct flock lock;
		ndbl->fd = db->fd(db);
		lock.l_type  = read_only ? F_RDLCK : F_WRLCK;
		lock.l_start = 0;
		lock.l_whence  = SEEK_SET;
		lock.l_len	 = 0;
		if(fcntl(ndbl->fd, F_SETLK, &lock))
			db_error();
	}
#endif	
	for(;;)
	{
		status = db->seq(db, &ky, &val, R_NEXT);
		if(status < 0)
			db_error();
		if(status)
			break;
		ndbl->reccount++;
	}
	all_tables = list_add(all_tables, ndbl);
	return ns;
}

storage *table_create(interpreter *i, char *name, char *arg)
{
	return itable_create(DB_BTREE, i, name, arg);
}

storage *hash_create(interpreter *i, char *name, char *arg)
{
	return itable_create(DB_HASH, i, name, arg);
}

static int find_table_in_list(dbf_list *w, dbf_list *key)
{
	return w == key;
}

extern int si_find_byso(shared_int *si, void *so);
static char *get_table_name(interpreter *i, storage *s)
{
	shared_int *si;
	if(!i)
		return "<unknown>";
	si = list_find(i->stos, s, (list_predicate_t)si_find_byso);
	return si ? si->intname : "<unknown>";
}

static void destroy_db(dbf_list *w)
{
#ifdef CAN_LOCK
	if(w->fd >= 0)
	{
		struct flock lock;
		lock.l_type = F_UNLCK;
		lock.l_start = 0;
		lock.l_whence = SEEK_SET;
		lock.l_len = 0;
		fcntl(w->fd, F_SETLKW, &lock);
	}
#endif	
	w->db->close(w->db);
}

void table_destroy(interpreter *i, storage *s)
{
	dbf_list *dbl = s->data;
	all_tables = list_remove(all_tables, dbl, (list_predicate_t)find_table_in_list,
				(list_destroy_t)destroy_db, sizeof(dbf_list)
				);
}

int table_check(interpreter *i, storage *s, dstring *key)
{
	dbdata_t ky;
	dbf_list *dbl = s->data;
	dbdata_t val;
	int status;
	
	ky.size = ds_length(key) + 1;
	ky.data = DS_BODY(key);
	status = dbl->db->get(dbl->db, &ky, &val, 0);
	if(status < 0)
		db_error();
	return !status;
}


dstring *table_get(interpreter *i, storage *s, dstring *key)
{
	dbf_list *dbl = s->data;
	dbdata_t ky, data;
	int status;

	if(!strcmp(DS_BODY(key), "__readonly__"))
		return ds_create(dbl->read_only ? "true" : NULL);
	if(!strcmp(DS_BODY(key), "__reversed__"))
		return ds_create(dbl->reversed ? "true" : NULL);
	ky.size = ds_length(key) + 1;
	ky.data = DS_BODY(key);
	status = dbl->db->get(dbl->db, &ky, &data, 0);
	if(status < 0)
		db_error();
	if(status)
		recover_error("No '%s' in table '%s'", DS_BODY(key), get_table_name(i, s));
	return ds_create(data.data);
}

void table_put(interpreter *i, storage *s, dstring *key, dstring *val)
{
	dbf_list *dbl = s->data;
	dbdata_t ky, tostore;
	int status;

	if(!strcmp(DS_BODY(key), "__reversed__"))
	{
		if(dbl->is_hash)
			recover_error("hashes cannot be scanned in a reverse order");
		dbl->reversed = !ds_isempty(val);
		return;
	}
	if(dbl->read_only)
		recover_error("attempt to modify a readonly table");
	if(!strcmp(DS_BODY(key), "__readonly__"))
	{
		dbl->read_only = !ds_isempty(val);
		return;
	}
	ky.size = ds_length(key) + 1;
	ky.data  = DS_BODY(key);
	status = dbl->db->get(dbl->db, &ky, &tostore, 0);
	if(status < 0)
		db_error();
	if(status)
		recover_error("No '%s' in table '%s'", DS_BODY(key), get_table_name(i, s));
	tostore.size = ds_length(val) + 1;
	tostore.data = DS_BODY(val);
	status = dbl->db->put(dbl->db, &ky, &tostore, 0);
	if(status < 0)
		db_error();
}

void table_add(interpreter *i, storage *s, dstring *key, dstring *val)
{
	dbdata_t ky;
	dbdata_t tostore;
	dbf_list *dbl = s->data;
	
	if(dbl->read_only)
		recover_error("attempt to modify a readonly table");
	ky.size = ds_length(key) + 1;
	ky.data = DS_BODY(key);
	tostore.size = ds_length(val) + 1;
	tostore.data = DS_BODY(val);
	if(dbl->db->put(dbl->db, &ky, &tostore, R_NOOVERWRITE))
		recover_error("'%s' already in table '%s'", DS_BODY(key), get_table_name(i, s));
	dbl->reccount++;
}

void table_remove(interpreter *i, storage *s, dstring *key)
{
	dbdata_t ky;
	dbf_list *dbl = s->data;
	if(dbl->read_only)
		recover_error("attempt to modify a readonly table");
	ky.size = ds_length(key) + 1;
	ky.data = DS_BODY(key);
	if(dbl->db->del(dbl->db, &ky, 0))
		recover_error("'%s' not in table '%s'", DS_BODY(key), get_table_name(i, s));
	dbl->reccount--;
}

dstring *table_first(interpreter *i, storage *s, void **iter)
{
	dbdata_t ky;
	dbdata_t val;
	dbf_list *dbl = s->data;
	int status;

	if(dbl->traverse_hash)
		recover_error("cannot do several traversals on hashes at once");
	status = dbl->db->seq(dbl->db, &ky, &val, dbl->reversed ? R_LAST : R_FIRST);
	if(status < 0)
		db_error();
	if(status)
		return NULL;
	if(dbl->is_hash)
		dbl->traverse_hash = 1;
	*iter = xstrdup(ky.data);
	return ds_create(ky.data);
}

dstring *table_next(interpreter *i, storage *s, void **iter)
{
	dbdata_t ky, val;
	int status;
	dbf_list *dbl = s->data;

	if(!dbl->is_hash)
	{
		ky.size = strlen(*iter) + 1;
		ky.data = *iter;
		status = dbl->db->seq(dbl->db, &ky, &val, R_CURSOR);
		if(status)
			db_error();
	}
	status = dbl->db->seq(dbl->db, &ky, &val, dbl->reversed ? R_PREV : R_NEXT);
	if(status < 0)
		db_error();
	free(*iter);
	if(status)
	{
		dbl->traverse_hash = 0;
		return *iter = NULL;
	}
	*iter = xstrdup(ky.data);
	return ds_create(*iter);
}

void table_enditer(interpreter *i, storage *s, void **iter)
{
	((dbf_list *)s->data)->traverse_hash = 0;
	if(*iter)
		free(*iter);
	*iter = NULL;
}

int table_numofitems(interpreter *i, storage *s)
{
	dbf_list *dbl = s->data;
	return dbl->reccount;
}

dstring *table_seek(interpreter *i, storage *s, void **iter, dstring *key)
{
	dbdata_t ky, val;
	int status;
	dbf_list *dbl = s->data;
	
	if(dbl->is_hash)
		recover_error("cannot traverse hashes from a random key");
	ky.size = ds_length(key) + 1;
	ky.data = DS_BODY(key);
	status = dbl->db->seq(dbl->db, &ky, &val, R_CURSOR);
	if(status < 0)
		db_error();
	*iter = status ? NULL : xstrdup(ky.data);
	return *iter ? ds_create(*iter) : NULL;
}

storage *table_clone(storage *src)
{
	recover_error("not implemented");
}

